/*
 * Author:	David Jones
 * Date:	1-JUN-1995
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <descrip.h>
#include "parse_log.h"

int parse_rec ( char *rec, char *out[7], int *status, int *bytes, int *date );

/*
 *   Open_log ( char *fname, since_date, before_date )
 *   read_log ( char *line, fields, req_status, bytes, date )
 *
 */
static FILE *logf;
static int since_date[6];
static int before_date[7];
int open_log ( char *fname, char *since, char *before )
{
    int i, status;
    char date_str[512];
    /*
     * Parse dates.
     */
    for ( i = 0; i < 6; i++ ) { since_date[i] = before_date[i] = 0; };
    before_date[0] = 5000;
    if ( since ) {
        sprintf ( date_str, "[%s]", since );
        status = parse_date ( date_str, since_date );
	if ( !status ) return status;
    }
    if ( before ) {
	sprintf ( date_str, "[%s]", before );
	status = parse_date ( date_str, before_date );
	if ( !status ) return status;
    }
    /*
     * Open global file pointer
     */
    logf = fopen ( fname, "r", "shr=upd", "mbc=64" , "dna=access.log" );
    if ( !logf ) return 0;
    return 1;
}
void close_log() { fclose(logf); }

static int compare_date ( int *date1, int *date2, int level )
{
    int i;
    for ( i = 0; i < level; i++ ) {
	if ( date1[i] < date2[i] ) return -1;
	else if ( date1[i] > date2[i] ) return 1;
    }
    return 0;
}

int read_log ( char *line, char *fields[7], int *req_status, int *bytes,
	int *date )
{
    static char cur_line[4096];
    int status;

    while ( fgets ( cur_line, sizeof(cur_line)-1, logf ) ) {
	if ( line ) strcpy ( line, cur_line );
	status = parse_rec ( cur_line, fields, req_status, bytes, date );
/* printf("read_log parse status: %d, date: %d %d %d\n", status,
date[0], date[1], date[2] ); */
	if ( !status ) continue;	/* ignore bad records */

	status = compare_date ( date, since_date, 6 );
/* printf("  compare since status: %d\n", status ); */
	if ( status < 0 ) continue;
	status = compare_date ( date, before_date, 6 );
/* printf("  compare before status: %d\n", status ); */
	if ( status  <= 0 ) return 1;
    }
    return 0;
}
/*
 *   0 - host
 *   1 - name
 *   2 - user
 *   3 - date
 *   4 - command
 *   5 - status
 *   6 - bytes
 */
int parse_rec ( char *rec, char *out[7], int *status, int *bytes, int *date )
{
    int state, i, val;
    char cur;

    *date = -1;
    for ( val = state = i = 0; cur = *rec; rec++ ) {
#ifdef DEBUG
printf("  state: %d cur: '%c' i: %d   val: %d\n", state, cur, i, val );
#endif
    switch ( state ) {
	case 0:			/* hostname */
	    out[0] = rec;
	    if ( cur != ' ' && cur != '\t' ) state = 1;
	    break;
	case 1:
	    if ( cur == ' ' || cur == '\t' ) { *rec = '\0'; state = 2; }
	    break;

	case 2:			/* remote user */
	    out[1] = rec;
	    if ( cur != ' ' && cur != '\t' ) state = 3;
	    break;
	case 3:
	    if ( cur == ' ' || cur == '\t' ) { *rec = '\0'; state = 4; }
	    break;

	case 4:			/* auth user */
	    out[2] = rec;
	    if ( cur != ' ' && cur != '\t' ) state = 5;
	    break;
	case 5:
	    if ( cur == ' ' || cur == '\t' ) { *rec = '\0'; state = 6; }
	    break;

	case 6:			/* date, '[' */
	    out[3] = rec;
	    if ( cur == '[' ) state = 9;
	    else if ( cur != ' ' && cur != '\t' ) state = 7;
	case 7:
		/* error state */
	    break;
	case 9:
	    if ( cur == ']' ) {
		if ( rec[1] == ' ' || rec[1] == '\t' ) { 
		    rec++;	*rec = '\0';
		    parse_date ( out[3], date );
		    state = 10; 
		} else state = 7;
	    }
	    break;

	case 10:		/* command state */
	    out[4] = rec;
	    if ( cur == '"' ) state = 11;
	    else if ( cur != ' ' && cur != '\t' ) state = 7;	/* error */
	    break;
	case 11:
	    if ( cur == '"' ) { 
		if ( rec[1] == ' ' || rec[1] == '\t' ) { 
		    rec++;	*rec = '\0'; 
		    state = 12; 
		} 
	    }
	    break;

	case 12:		/* status */
	    out[5] = rec;
	    if ( cur >= '0' && cur <= '9' ) {
		val = cur - '0';
		state = 13;
	    } else if ( cur != ' ' || cur != '\t' ) state = 7;
	    break;
	case 13:
	    if ( cur == ' ' || cur == '\t' ) { 
		*rec = '\0'; 
		*status = val;
		state = 14; 
	    }
	    else if ( cur < '0' || cur > '9' ) state = 7;
	    else {
		val = (val*10) + (cur - '0');
		if ( val >= 1000000000 ) state = 7;
	    }
	    break;

	case 14:		/* bytes */
	    out[6] = rec;
	    if ( cur >= '0' && cur <= '9' ) {
		state = 15;
		val = cur - '0';
	    } else if ( cur != ' ' || cur != '\t' ) state = 7;
	    break;
	case 15:
	    if ( cur == ' ' || cur == '\t' ||
		cur == '\n' ) { *rec = '\0'; *bytes = val; state = 16; }
	    else if ( cur < '0' || cur > '9' ) state = 7;
	    else {
		val = (val*10) + (cur - '0');
		if ( val >= 1000000000 ) state = 7;	/* too big */
	    }
	case 16:
	    break;
    }
    }
   return ( state > 14 ) ? 1 : 0;
}
/**************************************************************************/
/*  [dd/mon/yyyy:hh:mm:ss +offs]
 * date[0] - year
 * date[1] - month (1-12)
 * date[2] - day
 * date[3] - hour
 * date[4] - minute
 * date[5] - second
 */
int parse_date ( char *field, int *date )
{
   int state, i, val, j;
    static int month[12] = { 0 };
    static char *monthstr[12] = { "JAN", "FEB", "MAR","APR", "MAY", "JUN",
	"JUL", "AUG", "SEP", "OCT", "NOV", "DEC" };
    char cur;
    if ( month[0] == 0 ) {
	for ( i = 0; i < 12; i++ ) {
	    month[i] = (monthstr[i][0] * 1000000) + (monthstr[i][1]*1000) +
		monthstr[i][2];
	}
    }
    val = 0;
    for ( state = 0, i = 1; (cur = field[i]) != ']'; i++ ) switch ( state ) {
	case 0:
	    if ( cur == '/' ) {
		date[2] = val;
		val = 0;
		state = 1;
	    } else if ( (cur >= '0') && (cur <= '9') ) {
		val = (val*10) + (cur - '0');
	    } else if ( cur != ' ' && cur != '\t' ) return 0;
	    break;
	case 1:
	    if ( cur == '/' ) {
		for ( j = 0; j < 12; j++ ) if ( val == month[j] ) break;
		if ( j >= 12 ) return 0;	/* no match */
		date[1] = j + 1;
		val = 0;
		state = 2;
	    } else {
		if ( val >= 1000000 ) return 0;	/* too many chars */
		cur = _toupper(cur);
		val = val*1000 + cur;
	    }
	    break;

	case 2:
	    if ( cur == ':' || cur == ' ' ) {
		date[0] = val;
		date[3] = date[4] = date[5] = 0;
		val = 0;
		state = 3;
	    } else if ( cur >= '0' && cur <= '9' ) {
		val = (val*10) + (cur - '0');
	    } else return 0;
	    break;

	case 3:
	    if ( cur == ':' ) {
		date[3] = val;		/* hours */
		val = 0;
		state = 4;
	    } else if ( cur >= '0' && cur <= '9' ) {
		val = (val*10) + (cur - '0');
	    } else return 0;
	    break;

	case 4:
	    if ( cur == ':' ) {
		date[4] = val;		/* minutes */
		val = 0;
		state = 5;
	    } else if ( cur >= '0' && cur <= '9' ) {
		val = (val*10) + (cur - '0');
	    } else return 0;
	    break;

	case 5:
	    if ( cur == ' ' || cur == ']' || cur == '+' ) {
		date[5] = val;		/* seconds */
		val = 0;
		state = 6;
	    } else if ( cur >= '0' && cur <= '9' ) {
		val = (val*10) + (cur - '0');
	    } else return 0;
	    break;

	case 6:
	    break;	/* do nothing */
    }
    if ( state == 2 ) date[0] = val;
    return 1;
}

int julian_day ( int date[6] )
{
    int LIB$DAY(), SYS$BINTIM(), julian, status;
    long  *bintim, vms_time[2];
    char asctim[32];
    static $DESCRIPTOR(asctim_dx,"");
    static int prev_month=0, prev_year=0, prev_julian=0;
    static char *monthstr[12] = { "JAN", "FEB", "MAR","APR", "MAY", "JUN",
	"JUL", "AUG", "SEP", "OCT", "NOV", "DEC" };


    if ( date ) {
        if ( (prev_year == date[0]) && (prev_month == date[1]) ) {
	    julian = prev_julian + date[2];
	    return julian;
	}
        sprintf ( asctim, "%d-%s-%d 12:00:00", date[2], monthstr[date[1]-1],
		date[0] );

        asctim_dx.dsc$w_length = strlen(asctim);
        asctim_dx.dsc$a_pointer = asctim;
        status = SYS$BINTIM ( &asctim_dx, vms_time );
	bintim = vms_time;
    } else {
	/*
	 * Use current time, can't save it though.
	 */
	bintim = (long *) 0;
    }
    LIB$DAY ( &julian, bintim, 0 );

    if ( date ) {
        prev_year = date[0];
        prev_month = date[1];
        prev_julian = julian - date[2];
    }

    return julian;
}
